﻿#region License
/*
Kangaroo Library for C# and .NET
Copyright (c) 2013 Dimension Engineering LLC
http://www.dimensionengineering.com/kangaroo

Permission to use, copy, modify, and/or distribute this software for any
purpose with or without fee is hereby granted, provided that the above
copyright notice and this permission notice appear in all copies.

THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY
SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER
RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT,
NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE
USE OR PERFORMANCE OF THIS SOFTWARE.
*/
#endregion

namespace DimensionEngineering.Kangaroo
{
    /// <summary>
    /// Stores the response to a status request.
    /// </summary>
    public class KangarooStatus
    {
        static KangarooStatus _invalidStatus = CreateFromError(KangarooError.InvalidStatus);
        static KangarooStatus _portNotOpen = CreateFromError(KangarooError.PortNotOpen);
        static KangarooStatus _timedOut = CreateFromError(KangarooError.TimedOut);

        KangarooStatus()
        {
            Init();
        }

        /// <summary>
        /// Constructs a new KangarooStatus by parsing a status response.
        /// </summary>
        /// <param name="buffer">The buffer containing the status response to parse.</param>
        /// <param name="offset">The offset into the buffer.</param>
        /// <param name="count">The number of bytes.</param>
        public KangarooStatus(byte[] buffer, int offset, int count)
        {
            Init();
            if (Parse(buffer, offset, count)) { _valid = true; } else { Init(); }
        }

        void Init()
        {
            _valid = false;
            _channel = _flags = _echoCode = _sequenceCode = _type = 0;
            _value = 0;
        }

        bool Parse(byte[] buffer, int offset, int count)
        {
            KangarooReplyReader parser = new KangarooReplyReader(buffer, offset, count);

            if (!parser.TryRead(out _channel)) { return false; }
            if (!parser.TryRead(out _flags)) { return false; }
  
            if (0 != (_flags & (byte)KangarooStatusFlags.EchoCode))
            {
                if (!parser.TryRead(out _echoCode)) { return false; }
            }
  
            if (0 != (_flags & (byte)KangarooStatusFlags.SequenceCode))
            {
                if (!parser.TryRead(out _sequenceCode)) { return false; }
            }
    
            if (!parser.TryRead(out _type)) { return false; }
  
            _value = parser.ReadBitPackedNumber();
            if (0 != (_flags & (byte)KangarooStatusFlags.Error))
            {
                // Security check: Non-positive numbers should NEVER be received.
                //                 They are for library-returned errors (<0) and 'None' (=0).
                if (_value <= 0) { return false; }
            }

            return true;
        }

        internal static KangarooStatus CreateInvalidStatus()
        {
            return _invalidStatus;
        }

        internal static KangarooStatus CreatePortNotOpen()
        {
            return _portNotOpen;
        }

        internal static KangarooStatus CreateTimedOut()
        {
            return _timedOut;
        }

        static KangarooStatus CreateFromError(KangarooError error)
        {
            return new KangarooStatus()
            {
                _valid = true,
                _flags = (byte)KangarooStatusFlags.Error,
                _value = (int)error
            };
        }

        /// <summary>
        /// Gets the channel name.
        /// </summary>
        public char Channel
        {
            get { return (char)_channel; }    
        }

        /// <summary>
        /// Gets the echo code associated with this response.
        /// </summary>
        public byte EchoCode
        {
            get { return _echoCode; }
        }

        /// <summary>
        /// If the response is an error, gets the error.
        /// Otherwise, <see cref="KangarooError.None"/> is returned.
        /// </summary>
        public KangarooError Error
        {
            get { return IsError ? (KangarooError)Value : KangarooError.None; }
        }

        /// <summary>
        /// Gets the status flags associated with this response.
        /// </summary>
        public KangarooStatusFlags Flags
        {
            get { return (KangarooStatusFlags)_flags; }
        }

        /// <summary>
        /// Gets whether a recently-requested command is still executing.
        /// For example, <see cref="KangarooChannel.P"/> is busy while the channel is getting into position,
        /// and <see cref="KangarooChannel.Home"/> is busy while the channel is homing.
        /// Its opposite is <seealso cref="KangarooStatus.IsDone"/>.
        /// </summary>
        public bool IsBusy
        {
            get { return 0 != (Flags & KangarooStatusFlags.Busy); }
        }

        /// <summary>
        /// Gets whether a recently-requested command has completed execution.
        /// For example, <see cref="KangarooChannel.P"/> is done when the channel is in position.
        /// Its opposite is <seealso cref="KangarooStatus.IsBusy"/>.
        /// </summary>
        public bool IsDone
        {
            get { return !IsBusy; }
        }

        /// <summary>
        /// Returns true if the response is an error.
        /// Its opposite is <seealso cref="KangarooStatus.IsOK"/>.
        /// </summary>
        public bool IsError
        {
            get { return 0 != (Flags & KangarooStatusFlags.Error); }
        }

        /// <summary>
        /// Returns true if the response is not an error and <see cref="KangarooStatus.Value"/> is the requested value.
        /// Its opposite is <seealso cref="KangarooStatus.IsError"/>.
        /// </summary>
        public bool IsOK
        {
            get { return !IsError; }
        }

        /// <summary>
        /// Returns true if the KangarooStatus is valid.
        /// </summary>
        public bool IsValid
        {
            get { return _valid; }
        }

        /// <summary>
        /// Gets the sequence code from the most recent command.
        /// </summary>
        public byte SequenceCode
        {
            get { return _sequenceCode; }
        }

        /// <summary>
        /// Gets whether a timeout has occurred.
        /// </summary>
        public bool TimedOut
        {
            get { return Error == KangarooError.TimedOut; }
        }

        /// <summary>
        /// Gets the type of the request.
        /// </summary>
        public KangarooGetType Type
        {
            get { return (KangarooGetType)_type; }
        }

        /// <summary>
        /// Gets the value associated with this response.
        /// </summary>
        public int Value
        {
            get { return _value; }
        }

        bool _valid;
        byte _channel, _flags, _echoCode, _sequenceCode, _type;
        int _value;
    }
}
